@@ -1,3 +1,4 @@
+#![allow(clippy::type_complexity)]
 //! Lossless parser for deb822 style files.
 //!
 //! This parser can be used to parse files in the deb822 format, while preserving
@@ -159,6 +160,7 @@ fn parse(text: &str) -> Parse {
     impl Parser {
         fn parse_entry(&mut self) {
             self.builder.start_node(ENTRY.into());
+            // First, parse the key and colon
             if self.current() == Some(KEY) {
                 self.bump();
                 self.skip_ws();
@@ -191,10 +193,10 @@ fn parse(text: &str) -> Parse {
                     Some(NEWLINE) => {
                         self.bump();
                     }
-                    Some(_) => {
+                    Some(g) => {
                         self.builder.start_node(ERROR.into());
                         self.bump();
-                        self.errors.push("expected newline".to_string());
+                        self.errors.push(format!("expected newline, got {:?}", g));
                         self.builder.finish_node();
                     }
                 }
@@ -331,9 +333,9 @@ macro_rules! ast_node {
             }
         }
 
-        impl ToString for $ast {
-            fn to_string(&self) -> String {
-                self.0.text().to_string()
+        impl std::fmt::Display for $ast {
+            fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
+                write!(f, "{}", self.0.text())
             }
         }
     };
@@ -364,6 +366,96 @@ impl Deb822 {
         Deb822(SyntaxNode::new_root(builder.finish()).clone_for_update())
     }
 
+    /// Provide a formatter that can handle indentation and trailing separators
+    ///
+    /// # Arguments
+    /// * `control` - The control file to format
+    /// * `indentation` - The indentation to use
+    /// * `immediate_empty_line` - Whether the value should always start with an empty line. If true,
+    ///                  then the result becomes something like "Field:\n value". This parameter
+    ///                  only applies to the values that will be formatted over more than one line.
+    /// * `max_line_length_one_liner` - If set, then this is the max length of the value
+    ///                        if it is crammed into a "one-liner" value. If the value(s) fit into
+    ///                        one line, this parameter will overrule immediate_empty_line.
+    /// * `sort_paragraphs` - If set, then this function will sort the paragraphs according to the
+    ///                given function.
+    /// * `sort_entries` - If set, then this function will sort the entries according to the
+    ///               given function.
+    #[must_use]
+    pub fn wrap_and_sort(
+        self: &Deb822,
+        indentation: Indentation,
+        immediate_empty_line: bool,
+        max_line_length_one_liner: Option<usize>,
+        sort_paragraphs: Option<&dyn Fn(&Paragraph, &Paragraph) -> std::cmp::Ordering>,
+        sort_entries: Option<&dyn Fn(&Entry, &Entry) -> std::cmp::Ordering>,
+    ) -> Deb822 {
+        let mut builder = GreenNodeBuilder::new();
+        builder.start_node(ROOT.into());
+        let mut current = vec![];
+        let mut paragraphs = vec![];
+        for c in self.0.children_with_tokens() {
+            match c.kind() {
+                PARAGRAPH => {
+                    paragraphs.push((
+                        current,
+                        Paragraph::cast(c.as_node().unwrap().clone()).unwrap(),
+                    ));
+                    current = vec![];
+                }
+                COMMENT | ERROR => {
+                    current.push(c);
+                }
+                EMPTY_LINE => {
+                    current.extend(
+                        c.as_node()
+                            .unwrap()
+                            .children_with_tokens()
+                            .skip_while(|c| matches!(c.kind(), EMPTY_LINE | NEWLINE | WHITESPACE)),
+                    );
+                }
+                _ => {}
+            }
+        }
+        if let Some(sort_paragraph) = sort_paragraphs {
+            paragraphs.sort_by(|a, b| {
+                let a_key = &a.1;
+                let b_key = &b.1;
+                sort_paragraph(a_key, b_key)
+            });
+        }
+
+        for (i, mut paragraph) in paragraphs.into_iter().enumerate() {
+            if i > 0 {
+                builder.start_node(EMPTY_LINE.into());
+                builder.token(NEWLINE.into(), "\n");
+                builder.finish_node();
+            }
+            for c in paragraph.0.into_iter() {
+                builder.token(c.kind().into(), c.as_token().unwrap().text());
+            }
+            inject(
+                &mut builder,
+                paragraph
+                    .1
+                    .wrap_and_sort(
+                        indentation,
+                        immediate_empty_line,
+                        max_line_length_one_liner,
+                        sort_entries,
+                    )
+                    .0,
+            );
+        }
+
+        for c in current {
+            builder.token(c.kind().into(), c.as_token().unwrap().text());
+        }
+
+        builder.finish_node();
+        Self(SyntaxNode::new_root(builder.finish()).clone_for_update())
+    }
+
     /// Returns an iterator over all paragraphs in the file.
     pub fn paragraphs(&self) -> impl Iterator<Item = Paragraph> {
         self.0.children().filter_map(Paragraph::cast)
@@ -392,11 +484,13 @@ impl Deb822 {
         paragraph
     }
 
+    /// Read a deb822 file from the given path.
     pub fn from_file(path: impl AsRef<Path>) -> Result<Self, Error> {
         let text = std::fs::read_to_string(path)?;
         Ok(Self::from_str(&text)?)
     }
 
+    /// Read a deb822 file from the given path, ignoring any syntax errors.
     pub fn from_file_relaxed(
         path: impl AsRef<Path>,
     ) -> Result<(Self, Vec<String>), std::io::Error> {
@@ -406,8 +500,35 @@ impl Deb822 {
 
     pub fn from_str_relaxed(s: &str) -> (Self, Vec<String>) {
         let parsed = parse(s);
-        (parsed.root(), parsed.errors)
+        (parsed.root().clone_for_update(), parsed.errors)
     }
+
+    pub fn read<R: std::io::Read>(mut r: R) -> Result<Self, Error> {
+        let mut buf = String::new();
+        r.read_to_string(&mut buf)?;
+        Ok(Self::from_str(&buf)?)
+    }
+
+    pub fn read_relaxed<R: std::io::Read>(mut r: R) -> Result<(Self, Vec<String>), std::io::Error> {
+        let mut buf = String::new();
+        r.read_to_string(&mut buf)?;
+        Ok(Self::from_str_relaxed(&buf))
+    }
+}
+
+fn inject(builder: &mut GreenNodeBuilder, node: SyntaxNode) {
+    builder.start_node(node.kind().into());
+    for child in node.children_with_tokens() {
+        match child {
+            rowan::NodeOrToken::Node(child) => {
+                inject(builder, child);
+            }
+            rowan::NodeOrToken::Token(token) => {
+                builder.token(token.kind().into(), token.text());
+            }
+        }
+    }
+    builder.finish_node();
 }
 
 impl Paragraph {
@@ -419,6 +540,62 @@ impl Paragraph {
         Paragraph(SyntaxNode::new_root(builder.finish()).clone_for_update())
     }
 
+    #[must_use]
+    pub fn wrap_and_sort(
+        &mut self,
+        indentation: Indentation,
+        immediate_empty_line: bool,
+        max_line_length_one_liner: Option<usize>,
+        sort_entries: Option<&dyn Fn(&Entry, &Entry) -> std::cmp::Ordering>,
+    ) -> Paragraph {
+        let mut builder = GreenNodeBuilder::new();
+
+        let mut current = vec![];
+        let mut entries = vec![];
+
+        builder.start_node(PARAGRAPH.into());
+        for c in self.0.children_with_tokens() {
+            match c.kind() {
+                ENTRY => {
+                    entries.push((current, Entry::cast(c.as_node().unwrap().clone()).unwrap()));
+                    current = vec![];
+                }
+                ERROR | COMMENT => {
+                    current.push(c);
+                }
+                _ => {}
+            }
+        }
+
+        if let Some(sort_entry) = sort_entries {
+            entries.sort_by(|a, b| {
+                let a_key = &a.1;
+                let b_key = &b.1;
+                sort_entry(a_key, b_key)
+            });
+        }
+
+        for (pre, mut entry) in entries.into_iter() {
+            for c in pre.into_iter() {
+                builder.token(c.kind().into(), c.as_token().unwrap().text());
+            }
+
+            inject(
+                &mut builder,
+                entry
+                    .wrap_and_sort(indentation, immediate_empty_line, max_line_length_one_liner)
+                    .0,
+            );
+        }
+
+        for c in current {
+            builder.token(c.kind().into(), c.as_token().unwrap().text());
+        }
+
+        builder.finish_node();
+        Self(SyntaxNode::new_root(builder.finish()).clone_for_update())
+    }
+
     /// Returns the value of the given key in the paragraph.
     pub fn get(&self, key: &str) -> Option<String> {
         self.entries()
@@ -526,6 +703,39 @@ impl std::str::FromStr for Paragraph {
     }
 }
 
+#[cfg(feature = "python-debian")]
+impl pyo3::ToPyObject for Paragraph {
+    fn to_object(&self, py: pyo3::Python) -> pyo3::PyObject {
+        use pyo3::prelude::*;
+        let d = pyo3::types::PyDict::new_bound(py);
+        for (k, v) in self.items() {
+            d.set_item(k, v).unwrap();
+        }
+        let m = py.import_bound("debian.deb822").unwrap();
+        let cls = m.getattr("Deb822").unwrap();
+        cls.call1((d,)).unwrap().to_object(py)
+    }
+}
+
+#[cfg(feature = "python-debian")]
+impl pyo3::FromPyObject<'_> for Paragraph {
+    fn extract_bound(obj: &pyo3::Bound<pyo3::PyAny>) -> pyo3::PyResult<Self> {
+        use pyo3::prelude::*;
+        let d = obj.call_method0("__str__")?.extract::<String>()?;
+        Ok(Paragraph::from_str(&d)
+            .map_err(|e| pyo3::exceptions::PyValueError::new_err((e.to_string(),)))?)
+    }
+}
+
+#[derive(Debug, Clone, Copy, PartialEq, Eq)]
+pub enum Indentation {
+    /// Use the same indentation as the original line for the value.
+    FieldNameLength,
+
+    /// The number of spaces to use for indentation.
+    FixedIndentation(u32),
+}
+
 impl Entry {
     pub fn new(key: &str, value: &str) -> Entry {
         let mut builder = GreenNodeBuilder::new();
@@ -545,6 +755,105 @@ impl Entry {
         Entry(SyntaxNode::new_root(builder.finish()))
     }
 
+    #[must_use]
+    pub fn wrap_and_sort(
+        &mut self,
+        mut indentation: Indentation,
+        immediate_empty_line: bool,
+        max_line_length_one_liner: Option<usize>,
+    ) -> Entry {
+        let mut builder = GreenNodeBuilder::new();
+
+        let mut content = vec![];
+        builder.start_node(ENTRY.into());
+        for c in self.0.children_with_tokens() {
+            let text = c.as_token().map(|t| t.text());
+            match c.kind() {
+                KEY => {
+                    builder.token(KEY.into(), text.unwrap());
+                    if indentation == Indentation::FieldNameLength {
+                        indentation = Indentation::FixedIndentation(text.unwrap().len() as u32);
+                    }
+                }
+                COLON => {
+                    builder.token(COLON.into(), ":");
+                }
+                INDENT => {
+                    // Discard original whitespace
+                }
+                ERROR | COMMENT | VALUE | WHITESPACE | NEWLINE => {
+                    content.push(c);
+                }
+                EMPTY_LINE | ENTRY | ROOT | PARAGRAPH => unreachable!(),
+            }
+        }
+
+        let indentation = if let Indentation::FixedIndentation(i) = indentation {
+            i
+        } else {
+            1
+        };
+
+        assert!(indentation > 0);
+
+        // Strip trailing whitespace and newlines
+        while let Some(c) = content.last() {
+            if c.kind() == NEWLINE || c.kind() == WHITESPACE {
+                content.pop();
+            } else {
+                break;
+            }
+        }
+
+        let first_line_len = content
+            .iter()
+            .take_while(|c| c.kind() != NEWLINE)
+            .map(|c| c.as_token().unwrap().text().len())
+            .sum::<usize>() + self.key().map_or(0, |k| k.len()) + 2 /* ": " */;
+
+        let has_newline = content.iter().any(|c| c.kind() == NEWLINE);
+
+        let mut last_was_newline = false;
+        if max_line_length_one_liner
+            .map(|mll| first_line_len <= mll)
+            .unwrap_or(false)
+            && !has_newline
+        {
+            for c in content {
+                builder.token(c.kind().into(), c.as_token().unwrap().text());
+            }
+        } else {
+            if immediate_empty_line && has_newline {
+                builder.token(NEWLINE.into(), "\n");
+                last_was_newline = true;
+            } else {
+                builder.token(WHITESPACE.into(), " ");
+            }
+            // Strip leading whitespace and newlines
+            while let Some(c) = content.first() {
+                if c.kind() == NEWLINE || c.kind() == WHITESPACE {
+                    content.remove(0);
+                } else {
+                    break;
+                }
+            }
+            for c in content {
+                if last_was_newline {
+                    builder.token(INDENT.into(), &" ".repeat(indentation as usize));
+                }
+                builder.token(c.kind().into(), c.as_token().unwrap().text());
+                last_was_newline = c.kind() == NEWLINE;
+            }
+        }
+
+        if !last_was_newline {
+            builder.token(NEWLINE.into(), "\n");
+        }
+
+        builder.finish_node();
+        Self(SyntaxNode::new_root(builder.finish()).clone_for_update())
+    }
+
     pub fn key(&self) -> Option<String> {
         self.0
             .children_with_tokens()
@@ -954,10 +1263,11 @@ Foo: Blah
         use rowan::ast::AstNode;
 
         let entry = super::Entry::new("foo", "bar\nbaz");
-        let tokens: Vec<_> = entry.syntax()
-                                  .descendants_with_tokens()
-                                  .filter_map(|tok| tok.into_token())
-                                  .collect();
+        let tokens: Vec<_> = entry
+            .syntax()
+            .descendants_with_tokens()
+            .filter_map(|tok| tok.into_token())
+            .collect();
 
         assert_eq!("foo: bar\n baz\n", entry.to_string());
         assert_eq!("bar\nbaz", entry.value());
@@ -979,4 +1289,133 @@ Foo: Blah
                 .collect::<Vec<_>>()
         );
     }
+
+    #[test]
+    fn test_apt_entry() {
+        let text = r#"Package: cvsd
+Binary: cvsd
+Version: 1.0.24
+Maintainer: Arthur de Jong <adejong@debian.org>
+Build-Depends: debhelper (>= 9), po-debconf
+Architecture: any
+Standards-Version: 3.9.3
+Format: 3.0 (native)
+Files:
+ b7a7d67a02974c52c408fdb5e118406d 890 cvsd_1.0.24.dsc
+ b73ee40774c3086cb8490cdbb96ac883 258139 cvsd_1.0.24.tar.gz
+Vcs-Browser: http://arthurdejong.org/viewvc/cvsd/
+Vcs-Cvs: :pserver:anonymous@arthurdejong.org:/arthur/
+Checksums-Sha256:
+ a7bb7a3aacee19cd14ce5c26cb86e348b1608e6f1f6e97c6ea7c58efa440ac43 890 cvsd_1.0.24.dsc
+ 46bc517760c1070ae408693b89603986b53e6f068ae6bdc744e2e830e46b8cba 258139 cvsd_1.0.24.tar.gz
+Homepage: http://arthurdejong.org/cvsd/
+Package-List:
+ cvsd deb vcs optional
+Directory: pool/main/c/cvsd
+Priority: source
+Section: vcs
+
+"#;
+        let d: super::Deb822 = text.parse().unwrap();
+        let p = d.paragraphs().next().unwrap();
+        assert_eq!(p.get("Binary").as_deref(), Some("cvsd"));
+        assert_eq!(p.get("Version").as_deref(), Some("1.0.24"));
+        assert_eq!(
+            p.get("Maintainer").as_deref(),
+            Some("Arthur de Jong <adejong@debian.org>")
+        );
+    }
+
+    #[test]
+    fn test_format() {
+        let d: super::Deb822 = r#"Source: foo
+Maintainer: Foo Bar <foo@example.com>
+Section:      net
+Blah: blah  # comment
+Multi-Line:
+  Ahoi!
+     Matey!
+
+"#
+        .parse()
+        .unwrap();
+        let mut ps = d.paragraphs();
+        let mut p = ps.next().unwrap();
+        let result = p.wrap_and_sort(
+            super::Indentation::FieldNameLength,
+            false,
+            None,
+            None::<&dyn Fn(&super::Entry, &super::Entry) -> std::cmp::Ordering>,
+        );
+        assert_eq!(
+            result.to_string(),
+            r#"Source: foo
+Maintainer: Foo Bar <foo@example.com>
+Section: net
+Blah: blah  # comment
+Multi-Line: Ahoi!
+          Matey!
+"#
+        );
+    }
+
+    #[test]
+    fn test_format_sort_paragraphs() {
+        let d: super::Deb822 = r#"Source: foo
+Maintainer: Foo Bar <foo@example.com>
+
+# This is a comment
+Source: bar
+Maintainer: Bar Foo <bar@example.com>
+
+"#
+        .parse()
+        .unwrap();
+        let result = d.wrap_and_sort(
+            super::Indentation::FieldNameLength,
+            false,
+            None,
+            Some(&|a: &super::Paragraph, b: &super::Paragraph| {
+                a.get("Source").cmp(&b.get("Source"))
+            }),
+            None,
+        );
+        assert_eq!(
+            result.to_string(),
+            r#"# This is a comment
+Source: bar
+Maintainer: Bar Foo <bar@example.com>
+
+Source: foo
+Maintainer: Foo Bar <foo@example.com>
+"#,
+        );
+    }
+
+    #[test]
+    fn test_format_sort_fields() {
+        let d: super::Deb822 = r#"Source: foo
+Maintainer: Foo Bar <foo@example.com>
+Build-Depends: debhelper (>= 9), po-debconf
+Homepage: https://example.com/
+
+"#
+        .parse()
+        .unwrap();
+        let result = d.wrap_and_sort(
+            super::Indentation::FieldNameLength,
+            false,
+            None,
+            None,
+            Some(&|a: &super::Entry, b: &super::Entry| a.key().cmp(&b.key())),
+        );
+        assert_eq!(
+            result.to_string(),
+            r#"Build-Depends: debhelper (>= 9), po-debconf
+Homepage: https://example.com/
+Maintainer: Foo Bar <foo@example.com>
+Source: foo
+"#
+        );
+    }
 }
